import tensorflow as tf
import numpy as np
from tensorflow.keras import backend as K
from tensorflow.keras.initializers import Initializer
import skimage.morphology as skm
import scipy.ndimage.morphology as snm

MIN_LATT = -1
MAX_LATT = 0


class MinusOnesZeroCenter(Initializer):
    """Initializer that generates tensors initialized to -1 except for center value.

    Args:
        None

    Returns:
        tf.Tensor: Initialized tensor.
    """

    def __call__(self, shape, dtype=None):
        data = -np.ones(shape)
        data[int(shape[0] / 2), int(shape[1] / 2), :, :] = 0
        return tf.convert_to_tensor(data, np.float32)


class SparseZeros(Initializer):
    """Initializer that generates tensors initialized to MIN_LATT except for center value.

    Args:
        th (float): Threshold for generating sparse zeros.

    Returns:
        tf.Tensor: Initialized tensor.
    """

    def __init__(self, th=0.85):
        self.th = th

    def __call__(self, shape, dtype=None):
        data = np.random.random(shape)
        data = (data > self.th) * 1.0
        data = data + MIN_LATT
        return tf.convert_to_tensor(data, np.float32)

    def get_config(self):
        return {"th": self.th}


class SparseNumZeros(Initializer):
    """Initializer that generates tensors initialized to MIN_LATT except for center value.

    Args:
        th (int): Number of zeros to generate.

    Returns:
        tf.Tensor: Initialized tensor.
    """

    def __init__(self, th=0):
        self.th = th

    def __call__(self, shape, dtype=None):
        data = np.random.random(shape)
        v = np.sort(data.flatten())
        data = (data <= v[self.th]) * 1.0
        data = data + MIN_LATT
        return tf.convert_to_tensor(data, np.float32)

    def get_config(self):
        return {"th": self.th}


class SignedOnes(Initializer):
    """Initializer that generates tensors initialized to random -1 or 1 values.

    Args:
        minval (float): Lower bound of the range of random values.
        maxval (float): Upper bound of the range of random values.
        seed (int): Seed for the random generator.

    Returns:
        tf.Tensor: Initialized tensor.
    """

    def __init__(self, minval=MIN_LATT, maxval=-MIN_LATT, seed=None):
        self.minval = minval
        self.maxval = maxval
        self.seed = seed

    def __call__(self, shape, dtype=None):
        data = K.sign(K.random_uniform(shape, self.minval, self.maxval, dtype=dtype, seed=self.seed))
        if self.seed is not None:
            self.seed += 1
        return data

    def get_config(self):
        return {"minval": self.minval, "maxval": self.maxval, "seed": self.seed}


class MinusOnes(Initializer):
    """Initializer that generates tensors initialized to -1.

    Args:
        None

    Returns:
        tf.Tensor: Initialized tensor.
    """

    def __call__(self, shape, dtype=None):
        return K.constant(MIN_LATT, shape=shape, dtype=dtype)


class RandomLattice(Initializer):
    """Initializer that generates tensors with a uniform distribution (MIN_LATT,MAX_LATT).

    Args:
        minval (float): Lower bound of the range of random values.
        maxval (float): Upper bound of the range of random values.
        seed (int): Seed for the random generator.

    Returns:
        tf.Tensor: Initialized tensor.
    """

    def __init__(self, minval=MIN_LATT, maxval=MAX_LATT, seed=None):
        self.minval = minval
        self.maxval = maxval
        self.seed = seed

    def __call__(self, shape, dtype=None):
        data = K.random_uniform(shape, self.minval, self.maxval, dtype=dtype, seed=self.seed)
        if self.seed is not None:
            self.seed += 1
        return data

    def get_config(self):
        return {"minval": self.minval, "maxval": self.maxval, "seed": self.seed}


class RandomLatticewithZero(Initializer):
    """Initializer that generates tensors with a uniform distribution (MIN_LATT,-MIN_LATT).

    Args:
        minval (float): Lower bound of the range of random values.
        maxval (float): Upper bound of the range of random values.

    Returns:
        tf.Tensor: Initialized tensor.
    """

    def __init__(self, minval=MIN_LATT, maxval=MAX_LATT):
        self.minval = minval
        self.maxval = maxval

    def __call__(self, shape, dtype=None):
        data = K.random_uniform(shape, self.minval, self.maxval, dtype=dtype)
        data[int(shape[0] / 2), int(shape[1] / 2), :, :] = 0
        return tf.convert_to_tensor(data, np.float32)

    def get_config(self):
        return {"minval": self.minval, "maxval": self.maxval}


class Quadratic(Initializer):
    """Initializer with quadratic values.

    Args:
        tvalue (float): T-value for distance transform.
        cvalue (float): Constant multiplier.

    Returns:
        tf.Tensor: Initialized tensor.
    """

    def __init__(self, tvalue=2, cvalue=0.2):
        self.tvalue = tvalue
        self.cvalue = cvalue

    def __call__(self, shape, dtype=None):
        data = np.ones([shape[0], shape[1]])
        data[int(shape[0] / 2), int(shape[1] / 2)] = 0
        data = (snm.distance_transform_edt(data) / self.tvalue) ** 2
        data = -self.cvalue * (data)
        data = np.repeat(data[:, :, np.newaxis], shape[2], axis=2)
        data = np.repeat(data[:, :, :, np.newaxis], shape[3], axis=3)
        return tf.convert_to_tensor(data, np.float32)

    def get_config(self):
        return {"t_value": self.t_value, "c_value": self.c_value}


class SEinitializer(Initializer):
    """Initializer to a Structured Element (SE).

    Args:
        SE (np.ndarray): Structured Element.
        minval (float): Minimum value.

    Returns:
        tf.Tensor: Initialized tensor.
    """

    def __init__(self, SE=None, minval=None):
        self.SE = SE
        if minval is None:
            self.minval = MIN_LATT
        else:
            self.minval = minval

    def __call__(self, shape, dtype=None):
        data = np.zeros(shape)
        for i in range(data.shape[2]):
            for j in range(data.shape[3]):
                data[:, :, i, j] = self.SE + self.minval
        return tf.convert_to_tensor(data, np.float32)



import tensorflow as tf
import numpy as np
from tensorflow.keras.constraints import Constraint
from tensorflow.keras import backend as K
import skimage.morphology as skm
import scipy.ndimage.morphology as snm

MIN_LATT = -1
MAX_LATT = 0


@tf.custom_gradient
def rounding_op1(x):
    """Round the input tensor to the nearest tenth.

    Args:
        x (tf.Tensor): Input tensor.

    Returns:
        tf.Tensor: Rounded tensor.
    """

    def grad(dy):
        return dy

    return tf.round(x * 10) / 10, grad


@tf.custom_gradient
def rounding_op2(x):
    """Round the input tensor to the nearest hundredth.

    Args:
        x (tf.Tensor): Input tensor.

    Returns:
        tf.Tensor: Rounded tensor.
    """

    def grad(dy):
        return dy

    return tf.round(x * 100) / 100, grad


@tf.custom_gradient
def rounding_op3(x):
    """Round the input tensor to the nearest thousandth.

    Args:
        x (tf.Tensor): Input tensor.

    Returns:
        tf.Tensor: Rounded tensor.
    """

    def grad(dy):
        return dy

    return tf.round(x * 1000) / 1000, grad


@tf.custom_gradient
def rounding_op4(x):
    """Round the input tensor to the nearest ten-thousandth.

    Args:
        x (tf.Tensor): Input tensor.

    Returns:
        tf.Tensor: Rounded tensor.
    """

    def grad(dy):
        return dy

    return tf.round(x * 10000) / 10000, grad


class Rounding(Constraint):
    """Constrains weights by rounding them to specified decimal places.

    Args:
        c (int): Number of decimal places for rounding.
    """

    def __init__(self, c=4):
        self.c = c

    def __call__(self, w):
        if self.c == 1:
            return rounding_op1(w)
        elif self.c == 2:
            return rounding_op2(w)
        elif self.c == 3:
            return rounding_op3(w)
        else:
            # self.c==4:
            return rounding_op4(w)

    def get_config(self):
        return {"c": self.c}


class NonPositive(Constraint):
    """Constrains weights to be non-positive values."""

    def __init__(self):
        self.min_value = MIN_LATT
        self.max_value = MAX_LATT

    def __call__(self, w):
        return K.clip(w, self.min_value, self.max_value)

    def get_config(self):
        return {"min_value": self.min_value, "max_value": self.max_value}


class NonPositiveExtensive(Constraint):
    """Constrains weights to be non-positive and centers equal to zero."""

    def __init__(self):
        self.min_value = MIN_LATT
        self.max_value = MAX_LATT

    def __call__(self, w):
        w = K.clip(w, self.min_value, 0)
        data = np.ones(w.shape)
        data[int(w.shape[0] / 2), int(w.shape[1] / 2), :, :] = 0
        w = tf.multiply(w, tf.convert_to_tensor(data, np.float32))
        return w

    def get_config(self):
        return {"min_value": self.min_value, "max_value": self.max_value}


class ZeroToOne(Constraint):
    """Constrains weights to be between 0 and 1."""

    def __init__(self):
        self.min_value = 0.0
        self.max_value = 1.0

    def __call__(self, w):
        return K.clip(w, self.min_value, self.max_value)

    def get_config(self):
        return {"min_value": self.min_value, "max_value": self.max_value}


class Lattice(Constraint):
    """Constrains weights to be within a lattice range."""

    def __init__(self):
        self.min_value = MIN_LATT
        self.max_value = -MIN_LATT

    def __call__(self, w):
        w = K.clip(w, self.min_value, self.max_value)
        return w

    def get_config(self):
        return {"min_value": self.min_value, "max_value": self.max_value}


class SEconstraint(Constraint):
    """Constrains weights to any structured element (SE) shape."""

    def __init__(self, SE=skm.disk(1)):
        self.min_value = MIN_LATT
        self.max_value = -MIN_LATT
        self.data = SE

    def __call__(self, w):
        data = self.data
        data = np.repeat(data[:, :, np.newaxis], w.shape[2], axis=2)
        data = np.repeat(data[:, :, :, np.newaxis], w.shape[3], axis=3)
        w = w + (tf.convert_to_tensor(data, np.float32) + self.min_value)
        w = K.clip(w, self.min_value, self.max_value)
        return w

    def get_config(self):
        return {"min_value": self.min_value, "max_value": self.max_value, "SE": self.data}


class Disk(Constraint):
    """Constrains weights to a disk shape (only for square filters)."""

    def __init__(self):
        self.min_value = MIN_LATT
        self.max_value = -MIN_LATT

    def __call__(self, w):
        data = skm.disk(int(w.shape[0] / 2))
        data = np.repeat(data[:, :, np.newaxis], w.shape[2], axis=2)
        data = np.repeat(data[:, :, :, np.newaxis], w.shape[3], axis=3)
        w = w + (tf.convert_to_tensor(data, np.float32) + self.min_value)
        w = K.clip(w, self.min_value, self.max_value)
        return w

    def get_config(self):
        return {"min_value": self.min_value, "max_value": self.max_value}


class Diamond(Constraint):
    """Constrains weights to a diamond shape (only for square filters)."""

    def __init__(self):
        self.min_value = MIN_LATT
        self.max_value = -MIN_LATT

    def __call__(self, w):
        data = skm.diamond(int(w.shape[0] / 2))
        data = np.repeat(data[:, :, np.newaxis], w.shape[2], axis=2)
        data = np.repeat(data[:, :, :, np.newaxis], w.shape[3], axis=3)
        w = w + (tf.convert_to_tensor(data, np.float32) + self.min_value)
        w = K.clip(w, self.min_value, self.max_value)
        return w

    def get_config(self):
        return {"min_value": self.min_value, "max_value": self.max_value}
